/**
 * Hunter Adams (vha3@cornell.edu)
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 330 ohm resistor ---> VGA Green lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 19 ---> 220 ohm resistor ---> VGA Green hi_bit |   and to VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.1.1
 * Threads:
 * core 0:
 * histogram of random numbers
 * blink LED25 
 * core 1:
 * Toggle gpio 4 
 * Serial i/o 
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"
// // Our assembled programs:
// // Each gets the name <pio_filename.pio.h>
// #include "hsync.pio.h"
// #include "vsync.pio.h"
// #include "rgb.pio.h"

// ==========================================
// === protothreads setup
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "hardware/clocks.h"
#include "hardware/structs/clocks.h"
#include "hardware/regs/rosc.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"
// flag between serial and compute threads
int new_value ;
// number of rands to avareage to make gaussian
// and make one pixel worth more counts
int n_gauss, scale ;

// ==========================================
// === Ring Osc random bit RNG
// setup for higher speed oscillator
// ==========================================
void rosc_setup(void){
  volatile uint32_t *rosc_div = (uint32_t *)(ROSC_BASE + ROSC_DIV_OFFSET) ;
  volatile uint32_t *rosc_ctl = (uint32_t *)(ROSC_BASE + ROSC_CTRL_OFFSET) ;
  volatile uint32_t *rosc_freqA = (uint32_t *)(ROSC_BASE + ROSC_FREQA_OFFSET) ;
  volatile uint32_t *rosc_freqB = (uint32_t *)(ROSC_BASE + ROSC_FREQB_OFFSET) ;
  
  // set divider to one for frequency measurement
  *rosc_div = ROSC_DIV_VALUE_PASS + 1 ;
  // speed up the ROSC so more cycles between reads
  // (dont use ROSC_CTRL_FREQ_RANGE_VALUE_TOOHIGH)
  // Measured at 241 MHz with theses settings
  *rosc_ctl =  ROSC_CTRL_FREQ_RANGE_VALUE_HIGH ;// | ROSC_CTRL_ENABLE_VALUE_ENABLE;
  *rosc_freqA = (ROSC_FREQA_PASSWD_VALUE_PASS<<16) | 0xffff ;
  *rosc_freqB = (ROSC_FREQB_PASSWD_VALUE_PASS<<16) | 0xffff ;
}

// diagnostics
uint32_t rand_count, extractor_count;

// ==========================================================
// good quality 32-bit TRNG
// ==========================================================
// Excution time is 25-35 uSec, mostly becuause of the Von Neumann extactor
// This is going to be more useful for seed generation for rand()
// ==========================================================
// basic random number gen with Von Neumann extractor
// and a small delay in the extractor
// AND one pass through rand() function
// https://en.wikipedia.org/wiki/Randomness_extractor
uint32_t rand_rosc_VN(void){
    int k, random ;
    int random_bit1, random_bit2 ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    volatile uint32_t *rosc_ctl = (uint32_t *)(ROSC_BASE + ROSC_CTRL_OFFSET) ;
    
    for(k=0;k<32;k++){
      // von Neumann bit extractor         
      while(1){
        //extractor_count++ ;
        random_bit1=0x00000001 & (*rnd_reg);
        //  a small delay decorrelates the bits
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;
        asm("nop") ; asm("nop") ;      
        random_bit2=0x00000001 & (*rnd_reg);
        // if the two are diferent, use the first one
        if(random_bit1!=random_bit2) break;
      }  
      // build the 32 bit sample
      random=(random << 1) | random_bit1 ;
    }
    srand(random) ;
    //rand_count++ ;
    random = rand() ;
    return random;
}

// ==========================================================
//  medium quality 32-bit TRNG
// ==========================================================
// Excution time is 6 uSec
// Plenty good enough for game init with the fast RSOC clock
// ==========================================================
// and a small delay in the bit extractor
// AND one pass through rand() function
uint32_t rand_rosc(void){
    int k, random ;
    int random_bit1 ;
    volatile uint32_t *rnd_reg = (uint32_t *)(ROSC_BASE + ROSC_RANDOMBIT_OFFSET);
    
    for(k=0;k<32;k++){    
      // at least one nop need to be here for timing  
      //extractor_count++ ;
      asm("nop") ; asm("nop") ; // adds one microsec to execution
      //asm("nop") ; asm("nop") ;
      random_bit1=0x00000001 & (*rnd_reg);
      random=(random << 1) | random_bit1 ;
    }
    //rand_count++ ;
    srand(random) ;
    random = rand() ;
    return random;
}

//uint32_t lfsr_state = 0xdeadbeef ;
//uint32_t lfsr_reg ;


// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
short int histogram[512] ;
//
static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    static uint32_t rand_test ;
    static int begin_time, end_time ;
    
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // Draw some filled rectangles
    fillRect(64, 0, 176, 50, BLUE); // blue box
    fillRect(250, 0, 176, 50, RED); // red box
    fillRect(435, 0, 176, 50, GREEN); // green box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeString("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeString("Hardware ROSC random demo") ;

    setCursor(445, 10) ;
    setTextColor(BLACK) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.1.1") ;
    setCursor(445, 20) ;
    writeString("Mod for 16 colors (brl4)") ;
    setTextColor(WHITE) ;
    
    while(true) {
        // new_value set by serial thread
        PT_YIELD_UNTIL(pt, new_value==true) ;
        // erase old data
        fillRect(0, 70, 639, 480, BLACK); // black box
        memset(histogram, 0, sizeof(histogram)) ;
        rand_count = 0 ;
        //
        begin_time = PT_GET_TIME_usec() ;
        //
        while((histogram[rand_test]>>scale) <= 400){
            // VN version is slower, but better distribution
            //rand_test = rand_rosc_VN() & 0x01ff ;
            // call n times for gaussian
            //n_gauss = 6 ;
            rand_test = 0 ;
            for(int i=0; i<n_gauss; i++){
            rand_test += (rand_rosc()) & 0x01ff ;
            }
            rand_test = rand_test / n_gauss ;
            //
            rand_count++ ;
            histogram[rand_test]++ ;       
            drawPixel(50 + rand_test, 479-(histogram[rand_test]>>scale), MED_GREEN);
        }
      
        end_time = PT_GET_TIME_usec() - begin_time ;
        //printf("rand count=%d  usec=%d  usec/rand=%d\n", rand_count, end_time, end_time/rand_count) ;
        // fit gaussian
        // (1/(sigma*sqrt(2*pi))) exp[-0.5((x-mean)/sigma)^2] scaled to number of events
        // mean = 512*n_gauss/2   and sigma = sqrt(512*n_gauss/12)
        float x, y ;
        float sqrt2pi = sqrt(2*3.14159) ;
        float mean, sigma ;
        mean = 255.5 ; 
        sigma = 512/sqrt(12*n_gauss) ; // 44 n=12; 31 n=24; 62  n=6
        for(int i =0; i<512; i++){
            x = (float) i ;        
            y = (float)rand_count * exp(-0.5*(x-mean)*(x-mean)/(sigma*sigma)) / (sigma*sqrt2pi) ;
            //drawPixel(50 + (short)x, 479-(short)(y), RED);
            // fillCircle(short x0, short y0, short r, char color) ;
            fillCircle(50 + (short)x, 479-((short)(y)>>scale), 1, RED);
        }  
        // clear execute flag
        new_value = false ;
            
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 0
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "number_of_rand_in_gauss, number_of_counts/pixel: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
         sscanf(pt_serial_in_buffer, "%d %d", &n_gauss, &scale) ;
         
         // start comp thread
         new_value = true ;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // speed up the rosc
    rosc_setup();
    // First pass TRNG sanity check
    printf("first pass TRNG %08x\n", rand_rosc()) ;
    // get actual ROSC speed
    printf("ROSC freq=%d MHz\n", frequency_count_khz(CLOCKS_FC0_SRC_VALUE_ROSC_CLKSRC)/1000) ;

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  //multicore_reset_core1();
  //multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_serial) ;
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main